(import (srfi 130)
        (chicken base)
        (chicken condition)
        (chicken random)
        (only (chicken string) string-translate)
        (only (srfi 1) fold list-tabulate any every unfold take)
        test
        test-generative
        (prefix utf8 utf8:)
        (prefix utf8-srfi-13 utf8:)
        (only utf8-srfi-14 char-set:printing char-set-contains?))

;;;; Utility

;; Returns the number of elements of 'lis' that satisfy 'pred'.
(define (count pred lis)
  (fold (lambda (x total)
          (if (pred x)
              (+ total 1)
              total))
        0
        lis))

(define is-space? (cut char=? <> #\space))

(define (unicode-char-printing? c)
  (char-set-contains? char-set:printing c))

;;;; Test strings and string generators

(define test-string-max-length 64)

(define ascii "ABCDEF")

(define ABC "αβγ")

(define ABCDEF "ÀbÇdÉf")

;; List of length up to 'len-bound' of characters in the BMP.
(define (random-char-list len-bound)
  (list-tabulate
   (pseudo-random-integer len-bound)
   (lambda (junk)
     (integer->char (pseudo-random-integer #xffff)))))

;; Unicode string of length up to 'len-bound'.
(define (random-string len-bound)
  (utf8:list->string (random-char-list len-bound)))

;; String of fixed length with characters in the BMP.
(define (random-fixlen-string len)
  (utf8:list->string
   (list-tabulate
    len
    (lambda (junk)
      (integer->char (pseudo-random-integer #xffff))))))

;; Main randomized testing harness.
(define-syntax test-with-random-string
  (syntax-rules ()
    ((test-with-random-string var e0 e1 ...)
     (test-generative ((var (lambda ()
                              (random-string test-string-max-length))))
       e0 e1 ...))))

;; Aux. harness for fixed-length string tests.
(define-syntax test-with-random-fixlen-string
  (syntax-rules ()
    ((test-with-random-fixlen-string var len e0 e1 ...)
     (test-generative ((var (lambda ()
                              (random-fixlen-string len))))
       e0 e1 ...))))

;; Evaluates to true if the body expressions raises an object
;; satisfying pred, and false otherwise.
(define-syntax raises-exception/pred
  (syntax-rules ()
    ((raises-exception/pred pred e0 e1 ...)
     (handle-exceptions con (pred con) e0 e1 ...))))

;; Evaluates to true if the body expressions raise a type condition.
(define-syntax type-exception
  (syntax-rules ()
    ((type-exception e0 e1 ...)
     (raises-exception/pred
      (condition-predicate 'type)
      e0 e1 ...))))

;; Evaluates to true if the body expressions raises a valid exception.
;; (In particular, the raised object must be a condition.)
(define-syntax raises-exception
  (syntax-rules ()
    ((raises-exception e0 e1 ...)
     (raises-exception/pred condition? e0 e1 ...))))

;;;; Tests

(test-group "string-cursor?"
  (test #t (string-cursor? (string-index->cursor "" 0)))
  (test #t (not (string-cursor? #f)))
  (test-with-random-string s
    (test #t (string-cursor? (string-cursor-start s)))
    (test #t (string-cursor? (string-cursor-end s)))
    (test #t (string-cursor?
              (string-index->cursor s
                                    (quotient (utf8:string-length s) 2)))))
  (test #t (string-cursor? (string-index->cursor (make-string 10000) 9999)))
  )

(test-group "string-cursor-start"
  (test 0 (string-cursor->index "" (string-cursor-start "")))
  (test-with-random-string s
    (test 0 (string-cursor->index s (string-cursor-start s))))
  (test #t (type-exception (string-cursor-start #t)))
  )

(test-group "string-cursor-end"
  (test 0 (string-cursor->index "" (string-cursor-end "")))
  (test-with-random-string s
    (test (utf8:string-length s)
          (string-cursor->index s (string-cursor-end s))))
  (test #t (type-exception (string-cursor-start #t)))
  )

(test-group "string-cursor-next"
  (test-with-random-fixlen-string s 32
    (test 1 (string-cursor->index
             s
             (string-cursor-next s
                                 (string-cursor-start s))))
    (test 2 (string-cursor->index
             s
             (string-cursor-next
              s
              (string-cursor-next
               s
               (string-cursor-start s))))))
  (test #t (type-exception
            (string-cursor-next #t (string-cursor-start ""))))
  (test #t (type-exception (string-cursor-next ABC #t)))
  )

(test-group "string-cursor-prev"
  (test-with-random-fixlen-string s 32
    (test (- (utf8:string-length s) 1)
          (string-cursor->index
           s
           (string-cursor-prev s (string-cursor-end s))))
    (test (- (utf8:string-length s) 2)
          (string-cursor->index
           s
           (string-cursor-prev
            s
            (string-cursor-prev
             s
             (string-cursor-end s))))))
  (test #t (type-exception
            (string-cursor-prev #t (string-cursor-start ""))))
  (test #t (type-exception (string-cursor-prev ABC #t)))
  )

(test-group "string-cursor-forward"
  (test-with-random-fixlen-string s 32
    (test 0
          (string-cursor->index
           s
           (string-cursor-forward s (string-cursor-start s) 0)))
    (test 2
          (string-cursor->index
           s
           (string-cursor-forward s (string-cursor-start s) 2))))
  (test #t (type-exception
            (string-cursor-forward #t (string-cursor-start "") 0)))
  (test #t (type-exception (string-cursor-forward ABC #t 0)))
  (test #t (type-exception
            (string-cursor-forward ABC (string-cursor-start ABC) #t)))
  )

(test-group "string-cursor-back"
  (test-with-random-fixlen-string s 32
    (test (utf8:string-length s)
          (string-cursor->index
           s
           (string-cursor-back s (string-cursor-end s) 0)))
    (test (- (utf8:string-length s) 2)
          (string-cursor->index
           s
           (string-cursor-back s (string-cursor-end s) 2))))
  (test #t (type-exception
            (string-cursor-back #t (string-cursor-start "") 0)))
  (test #t (type-exception (string-cursor-back ABC #t 0)))
  (test #t (type-exception
            (string-cursor-back ABC (string-cursor-start ABC) #t)))
  )

(test-group "string-cursor=?"
  (test-with-random-fixlen-string s 4
    (let ((start (string-cursor-start s)))
      (test #t (string-cursor=? start start))
      (test #f (string-cursor=? start (string-cursor-next s start)))
      (test #f (string-cursor=? start (string-cursor-forward s start 2)))
      (test #f (string-cursor=? start (string-cursor-forward s start 3)))
      (test #f (string-cursor=? (string-cursor-next s start) start))
      (test #t (string-cursor=? (string-cursor-next s start)
                                (string-cursor-next s start)))
      (test #f (string-cursor=? (string-cursor-next s start)
                                (string-cursor-forward s start 2)))
      (test #f (string-cursor=? (string-cursor-next s start)
                                (string-cursor-forward s start 3)))
      (test #f (string-cursor=? (string-cursor-forward s start 2) start))
      (test #f (string-cursor=? (string-cursor-forward s start 2)
                                (string-cursor-next s start)))
      (test #t (string-cursor=? (string-cursor-forward s start 2)
                                (string-cursor-forward s start 2)))
      (test #f (string-cursor=? (string-cursor-forward s start 2)
                                (string-cursor-forward s start 3)))
      (test #f (string-cursor=? (string-cursor-forward s start 3) start))
      (test #f (string-cursor=? (string-cursor-forward s start 3)
                                (string-cursor-next s start)))
      (test #f (string-cursor=? (string-cursor-forward s start 3)
                                (string-cursor-forward s start 2)))
      (test #t (string-cursor=? (string-cursor-forward s start 3)
                                (string-cursor-forward s start 3)))
      ))
  (test #t (type-exception (string-cursor=? #t (string-cursor-start ABC))))
  (test #t (type-exception (string-cursor=? (string-cursor-start ABC) #t)))
  )

(test-group "string-cursor<?"
  (test-with-random-fixlen-string s 4
    (let ((start (string-cursor-start s)))
      (test #f (string-cursor<? start start))
      (test #t (string-cursor<? start (string-cursor-next s start)))
      (test #t (string-cursor<? start (string-cursor-forward s start 2)))
      (test #t (string-cursor<? start (string-cursor-forward s start 3)))
      (test #f (string-cursor<? (string-cursor-next s start) start))
      (test #f (string-cursor<? (string-cursor-next s start)
                                (string-cursor-next s start)))
      (test #t (string-cursor<? (string-cursor-next s start)
                                (string-cursor-forward s start 2)))
      (test #t (string-cursor<? (string-cursor-next s start)
                                (string-cursor-forward s start 3)))
      (test #f (string-cursor<? (string-cursor-forward s start 2) start))
      (test #f (string-cursor<? (string-cursor-forward s start 2)
                                (string-cursor-next s start)))
      (test #f (string-cursor<? (string-cursor-forward s start 2)
                                (string-cursor-forward s start 2)))
      (test #t (string-cursor<? (string-cursor-forward s start 2)
                                (string-cursor-forward s start 3)))
      (test #f (string-cursor<? (string-cursor-forward s start 3) start))
      (test #f (string-cursor<? (string-cursor-forward s start 3)
                                (string-cursor-next s start)))
      (test #f (string-cursor<? (string-cursor-forward s start 3)
                                (string-cursor-forward s start 2)))
      (test #f (string-cursor<? (string-cursor-forward s start 3)
                                (string-cursor-forward s start 3)))
      ))
  (test #t (type-exception (string-cursor<? #t (string-cursor-start ABC))))
  (test #t (type-exception (string-cursor<? (string-cursor-start ABC) #t)))
  )

(test-group "string-cursor>?"
  (test-with-random-fixlen-string s 4
    (let ((start (string-cursor-start s)))
      (test #f (string-cursor>? start start))
      (test #f (string-cursor>? start (string-cursor-next s start)))
      (test #f (string-cursor>? start (string-cursor-forward s start 2)))
      (test #f (string-cursor>? start (string-cursor-forward s start 3)))
      (test #t (string-cursor>? (string-cursor-next s start) start))
      (test #f (string-cursor>? (string-cursor-next s start)
                                (string-cursor-next s start)))
      (test #f (string-cursor>? (string-cursor-next s start)
                                (string-cursor-forward s start 2)))
      (test #f (string-cursor>? (string-cursor-next s start)
                                (string-cursor-forward s start 3)))
      (test #t (string-cursor>? (string-cursor-forward s start 2) start))
      (test #t (string-cursor>? (string-cursor-forward s start 2)
                                (string-cursor-next s start)))
      (test #f (string-cursor>? (string-cursor-forward s start 2)
                                (string-cursor-forward s start 2)))
      (test #f (string-cursor>? (string-cursor-forward s start 2)
                                (string-cursor-forward s start 3)))
      (test #t (string-cursor>? (string-cursor-forward s start 3) start))
      (test #t (string-cursor>? (string-cursor-forward s start 3)
                                (string-cursor-next s start)))
      (test #t (string-cursor>? (string-cursor-forward s start 3)
                                (string-cursor-forward s start 2)))
      (test #f (string-cursor>? (string-cursor-forward s start 3)
                                (string-cursor-forward s start 3)))
      ))
  (test #t (type-exception (string-cursor>? #t (string-cursor-start ABC))))
  (test #t (type-exception (string-cursor>? (string-cursor-start ABC) #t)))
  )

(test-group "string-cursor<=?"
  (test-with-random-fixlen-string s 4
    (let ((start (string-cursor-start s)))
      (test #t (string-cursor<=? start start))
      (test #t (string-cursor<=? start (string-cursor-next s start)))
      (test #t (string-cursor<=? start (string-cursor-forward s start 2)))
      (test #t (string-cursor<=? start (string-cursor-forward s start 3)))
      (test #f (string-cursor<=? (string-cursor-next s start) start))
      (test #t (string-cursor<=? (string-cursor-next s start)
                                 (string-cursor-next s start)))
      (test #t (string-cursor<=? (string-cursor-next s start)
                                 (string-cursor-forward s start 2)))
      (test #t (string-cursor<=? (string-cursor-next s start)
                                 (string-cursor-forward s start 3)))
      (test #f (string-cursor<=? (string-cursor-forward s start 2) start))
      (test #f (string-cursor<=? (string-cursor-forward s start 2)
                                 (string-cursor-next s start)))
      (test #t (string-cursor<=? (string-cursor-forward s start 2)
                                 (string-cursor-forward s start 2)))
      (test #t (string-cursor<=? (string-cursor-forward s start 2)
                                 (string-cursor-forward s start 3)))
      (test #f (string-cursor<=? (string-cursor-forward s start 3) start))
      (test #f (string-cursor<=? (string-cursor-forward s start 3)
                                 (string-cursor-next s start)))
      (test #f (string-cursor<=? (string-cursor-forward s start 3)
                                 (string-cursor-forward s start 2)))
      (test #t (string-cursor<=? (string-cursor-forward s start 3)
                                 (string-cursor-forward s start 3)))
      ))
  (test #t (type-exception (string-cursor<=? #t (string-cursor-start ABC))))
  (test #t (type-exception (string-cursor<=? (string-cursor-start ABC) #t)))
  )

(test-group "string-cursor>=?"
  (test-with-random-fixlen-string s 4
    (let ((start (string-cursor-start s)))
      (test #t (string-cursor>=? start start))
      (test #f (string-cursor>=? start (string-cursor-next s start)))
      (test #f (string-cursor>=? start (string-cursor-forward s start 2)))
      (test #f (string-cursor>=? start (string-cursor-forward s start 3)))
      (test #t (string-cursor>=? (string-cursor-next s start) start))
      (test #t (string-cursor>=? (string-cursor-next s start)
                                 (string-cursor-next s start)))
      (test #f (string-cursor>=? (string-cursor-next s start)
                                 (string-cursor-forward s start 2)))
      (test #f (string-cursor>=? (string-cursor-next s start)
                                 (string-cursor-forward s start 3)))
      (test #t (string-cursor>=? (string-cursor-forward s start 2) start))
      (test #t (string-cursor>=? (string-cursor-forward s start 2)
                                 (string-cursor-next s start)))
      (test #t (string-cursor>=? (string-cursor-forward s start 2)
                                 (string-cursor-forward s start 2)))
      (test #f (string-cursor>=? (string-cursor-forward s start 2)
                                 (string-cursor-forward s start 3)))
      (test #t (string-cursor>=? (string-cursor-forward s start 3) start))
      (test #t (string-cursor>=? (string-cursor-forward s start 3)
                                 (string-cursor-next s start)))
      (test #t (string-cursor>=? (string-cursor-forward s start 3)
                                 (string-cursor-forward s start 2)))
      (test #t (string-cursor>=? (string-cursor-forward s start 3)
                                 (string-cursor-forward s start 3)))
      ))
  (test #t (type-exception (string-cursor>=? #t (string-cursor-start ABC))))
  (test #t (type-exception (string-cursor>=? (string-cursor-start ABC) #t)))
  )

(test-group "string-cursor-diff"
  (test 0 (string-cursor-diff ""
                              (string-cursor-start "")
                              (string-cursor-start "")))
  (test-with-random-string s
    (let* ((s-len (utf8:string-length s))
           (k (quotient s-len 2)))
      (test s-len (string-cursor-diff s
                                      (string-cursor-start s)
                                      (string-cursor-end s)))
      (test k (string-cursor-diff s
                                  (string-cursor-start s)
                                  (string-index->cursor s k)))
      (test (- s-len k) (string-cursor-diff s
                                            (string-index->cursor s k)
                                            (string-cursor-end s)))
      ))
  (test #t (type-exception
            (string-cursor-diff #t
                                (string-cursor-start ABC)
                                (string-cursor-start ABC))))
  (test #t (type-exception
            (string-cursor-diff ABC #t (string-cursor-start ABC))))
  (test #t (type-exception
            (string-cursor-diff ABC (string-cursor-start ABC) #t)))
  )

(test-group "string-cursor->index"
  (test-with-random-fixlen-string s 4
    (test 0 (string-cursor->index s (string-index->cursor s 0)))
    (test 3 (string-cursor->index s (string-index->cursor s 3)))
    )
  (test #t (type-exception
            (string-cursor->index #t (string-cursor-start ABC))))
  (test #t (type-exception (string-cursor->index ABC #t)))
  )

(test-group "string-index->cursor"
  (test-with-random-fixlen-string s 4
    (test #t
          (string-cursor=?
           (string-cursor-start s)
           (string-index->cursor
            s
            (string-cursor->index s (string-cursor-start s)))))
    (test #t
          (string-cursor=?
           (string-cursor-back s (string-cursor-end s) 1)
           (string-index->cursor
            s
            (string-cursor->index
             s
             (string-cursor-back s (string-cursor-end s) 1)))))
    )
  (test #t (type-exception (string-index->cursor #t 0)))
  (test #t (type-exception (string-index->cursor ABC #t)))
  )

;;;; Predicates

(test-group "string-null?"
  (test #t (string-null? ""))
  (test #f (string-null? "abc"))
  (test #t (type-exception (string-null? 0)))
  )

(test-group "string-every"
  ;;; Need some Unicode tests.

  (test #t (string-every (lambda (c) (if (char? c) c #f)) ""))
  (test #\c (string-every (lambda (c) (if (char? c) c #f)) "abc"))
  (test #f (string-every (lambda (c) (if (char>? c #\b) c #f)) "abc"))
  (test #\c (string-every (lambda (c) (if (char>? c #\b) c #f))
                          "abc"
                          (string-index->cursor "abc" 2)))
  (test #t (string-every (lambda (c) (if (char>? c #\b) c #f))
                         "abc"
                         (string-index->cursor "abc" 1)
                         (string-index->cursor "abc" 1)))
  (test-with-random-string s
    (let ((print? (lambda (c) (and (unicode-char-printing? c) c)))
          (lis (utf8:string->list s)))
      (test #t (string-every (lambda (c)
                               (< (char->integer c) #xffff))
                             s))
      (test (every print? lis) (string-every print? s))))

  (test #t (type-exception (string-every 0 "")))
  (test #t (type-exception (string-every char? 0)))
  (test #t (type-exception (string-every char? "" #t)))
  (test #t (type-exception (string-every char? "" 0 #t)))
  )

(test-group "string-any"
  ;;; Need some Unicode tests.

  (test #f (string-any (lambda (c) (if (char? c) c #f)) ""))
  (test #\a (string-any (lambda (c) (if (char? c) c #f)) "abc"))
  (test #\c (string-any (lambda (c) (if (char>? c #\b) c #f)) "abc"))
  (test #\c (string-any (lambda (c) (if (char>? c #\b) c #f))
                        "abc"
                        (string-index->cursor "abc" 2)))
  (test #f (string-any (lambda (c) (if (char>? c #\b) c #f))
                       "abc"
                       (string-cursor-start "abc")
                       (string-index->cursor "abc" 2)))

  (test-with-random-string s
    (let ((print? (lambda (c) (and (unicode-char-printing? c) c)))
          (lis (utf8:string->list s)))
      (test #f (string-any (lambda (c)
                             (>= (char->integer c) #xffff))
                           s))
      (test (any print? lis) (string-any print? s))))

  (test #t (type-exception (string-any 0 "")))
  (test #t (type-exception (string-any char? 0)))
  (test #t (type-exception (string-any char? "" #t)))
  (test #t (type-exception (string-any char? "" 0 #t)))
  )

;;;; Constructors

(test-group "string-tabulate"
  (test ""
        (string-tabulate (lambda (i)
                           (integer->char (+ i (char->integer #\a))))
                         0))
  (test-with-random-string s
    (test s
          (string-tabulate (cut utf8:string-ref s <>)
                           (utf8:string-length s))))
  (test #t (type-exception (string-tabulate 0 8)))
  (test #t (type-exception (string-tabulate integer->char 8.2)))
  )

(test-group "string-unfold"
  (test "" (string-unfold null? car cdr '()))
  (test ABC (string-unfold null? car cdr (utf8:string->list ABC)))
  (test-with-random-string s
    (let ((start (string-cursor-start s))
          (end (string-cursor-end s))
          (s-rev (utf8:string-reverse s)))
      (test s
            (string-unfold (cut string-cursor=? <> end)
                           (cut string-ref/cursor s <>)
                           (cut string-cursor-next s <>)
                           start))
      (test (string-append ABC s ABCDEF)
            (string-unfold (cut string-cursor=? <> end)
                           (cut string-ref/cursor s <>)
                           (cut string-cursor-next s <>)
                           start
                           ABC
                           (constantly ABCDEF)))
    ))
  (test #t (type-exception (string-unfold 0 values values #t)))
  (test #t (type-exception (string-unfold values 0 values #t)))
  (test #t (type-exception (string-unfold values values 0 #t)))
  (test #t (type-exception (string-unfold values values values #t 0)))
  (test #t (type-exception (string-unfold values values values #t "" 0)))
  )

(test-group "string-unfold-right"
  (test "" (string-unfold-right null? car cdr '()))
  (test (utf8:string-reverse ABC)
        (string-unfold-right null? car cdr (utf8:string->list ABC)))
  (test-with-random-string s
    (let ((start (string-cursor-start s))
          (end (string-cursor-end s))
          (s-rev (utf8:string-reverse s)))
      (test s-rev
            (string-unfold-right (cut string-cursor=? <> end)
	                         (cut string-ref/cursor s <>)
                                 (cut string-cursor-next s <>)
                                 start))
      (test (string-append ABCDEF s-rev ABC)
            (string-unfold-right (cut string-cursor=? <> end)
                                 (cut string-ref/cursor s <>)
                                 (cut string-cursor-next s <>)
                                 start
                                 ABC
                                 (constantly ABCDEF)))
    ))
  (test #t (type-exception (string-unfold-right 0 values values #t)))
  (test #t (type-exception (string-unfold-right values 0 values #t)))
  (test #t (type-exception (string-unfold-right values values 0 #t)))
  (test #t (type-exception (string-unfold-right values values values #t 0)))
  (test #t (type-exception (string-unfold-right values values values #t "" 0)))
  )

;;;; Conversion

(test-group "string->list/cursors"
  (test '() (string->list/cursors ""))
  (test '() (string->list/cursors "" 0))
  (test '() (string->list/cursors "" 0 0))
  (test (map integer->char '(#x3b1 #x3b2 #x3b3))
        (string->list/cursors ABC))
  (test (map integer->char '(#x3b2 #x3b3))
        (string->list/cursors ABC 1))
  (test (map integer->char '(#x3b2))
        (string->list/cursors ABC 1 2))

  ;; I assume the utf8 egg got this one right...
  (test-with-random-fixlen-string s 32
    (test (utf8:string->list s) (string->list/cursors s))
    (test (utf8:string->list s 8 16)
          (string->list/cursors s
                                (string-index->cursor s 8)
                                (string-index->cursor s 16)))
    )
  (test #t (type-exception (string->list/cursors 0)))
  (test #t (type-exception (string->list/cursors "" #t)))
  (test #t (type-exception (string->list/cursors "" 0 #t)))
  )

(test-group "string->vector/cursors"
  (test '#() (string->vector/cursors ""))
  (test '#() (string->vector/cursors "" 0))
  (test '#() (string->vector/cursors "" 0 0))
  (test (list->vector (map integer->char '(#x3b1 #x3b2 #x3b3)))
        (string->vector/cursors ABC))
  (test (list->vector (map integer->char '(#x3b2 #x3b3)))
        (string->vector/cursors ABC 1))
  (test (list->vector (map integer->char '(#x3b2)))
        (string->vector/cursors ABC 1 2))
  (test #t (type-exception (string->vector/cursors 0)))
  (test #t (type-exception (string->vector/cursors "" #t)))
  (test #t (type-exception (string->vector/cursors "" 0 #t)))
  )

(test-group "reverse-list->string"
  (test "" (reverse-list->string '()))
  (test "a" (reverse-list->string '(#\a)))
  (test-generative ((lis (lambda () (random-char-list 64))))
    (test (utf8:list->string (reverse lis))
          (reverse-list->string lis))
    )
  (test #t (type-exception (reverse-list->string #t)))
  )

(test-group "string-join"
  (test "" (string-join '()))
  (test "" (string-join '() "α"))
  (test "" (string-join '() "α" 'infix))
  (test "" (string-join '() "α" 'prefix))
  (test "" (string-join '() "α" 'suffix))
  (test "" (string-join '("")))
  (test "" (string-join '("") "α"))
  (test "" (string-join '("") "α" 'infix))
  (test "α" (string-join '("") "α" 'prefix))
  (test "α" (string-join '("") "α" 'suffix))
  (test #t (raises-exception (string-join '() "α" 'strict-infix)))
  (test #t (raises-exception (string-join '() "α" 'affix)))
  
  (test-generative ((s1 (lambda ()
                          (random-string test-string-max-length)))
                    (s2 (lambda ()
                          (random-string test-string-max-length))))
    (test s1 (string-join (list s1)))
    (test (string-append s1 " " s2) (string-join (list s1 s2)))
    (test (string-append s1 "α" s2) (string-join (list s1 s2) "α"))
    (test (string-append s1 "α" s2) (string-join (list s1 s2) "α" 'infix))
    (test (string-append s1 "α" s2)
          (string-join (list s1 s2) "α" 'strict-infix))
    (test (string-append "α" s1 "α" s2)
          (string-join (list s1 s2) "α" 'prefix))
    (test (string-append s1 "α" s2 "α")
          (string-join (list s1 s2) "α" 'suffix))
    )
  (test #t (type-exception (string-join #t)))
  (test #t (type-exception (string-join '() #t)))
  (test #t (type-exception (string-join '() "" #t)))
  )

;;;; Selection

(test-group "string-ref/cursor"
  (let ((ascii-start (string-cursor-start ascii)))
    (test #\A (string-ref/cursor ascii ascii-start))
    (test #\B (string-ref/cursor ascii
                                 (string-cursor-next ascii ascii-start)))
    (test #\C (string-ref/cursor ascii
                                 (string-cursor-forward ascii
                                                        ascii-start 2))))

  (let ((ABC-start (string-cursor-start ABC)))
    (test "α" (utf8:string (string-ref/cursor ABC ABC-start)))
    (test "β" (utf8:string
               (string-ref/cursor ABC
                                  (string-cursor-next ABC ABC-start))))
    (test "γ" (utf8:string
               (string-ref/cursor ABC
                                  (string-cursor-forward
                                   ABC
                                   ABC-start
                                   2)))))

  (test-with-random-string s
    (let ((end (string-cursor-end s)))
      (test (utf8:string->list s)
            (unfold (cut string-cursor>=? <> end)
                    (cut string-ref/cursor s <>)
                    (cut string-cursor-next s <>)
                    (string-cursor-start s)))
    ))
  (test #t (type-exception (string-ref/cursor 0 0)))
  (test #t (type-exception (string-ref/cursor "" #t)))
  )

(test-group "substring/cursors"
  (test "" (substring/cursors ""
                              (string-cursor-start "")
                              (string-cursor-start "")))
  (test "" (substring/cursors ascii
                              (string-cursor-start ascii)
                              (string-cursor-start ascii)))
  (test "" (substring/cursors ascii
                              (string-cursor-end ascii)
                              (string-cursor-end ascii)))
  (test ascii (substring/cursors ascii
                                 (string-cursor-start ascii)
                                 (string-cursor-end ascii)))
  (test ABC (substring/cursors ABC
                               (string-cursor-start ABC)
                               (string-cursor-end ABC)))
  (test ABCDEF (substring/cursors ABCDEF
                                  (string-cursor-start ABCDEF)
                                  (string-cursor-end ABCDEF)))
  (test "αβ"
        (substring/cursors ABC
                           (string-cursor-start ABC)
                           (string-cursor-prev
                            ABC
                            (string-cursor-end ABC))))
  (test "βγ"
        (substring/cursors ABC
                           (string-cursor-next
                            ABC
                            (string-cursor-start ABC))
                           (string-cursor-end ABC)))
  (test "β"
        (substring/cursors ABC
                           (string-cursor-next
                            ABC
                            (string-cursor-start ABC))
                           (string-cursor-prev
                            ABC
                            (string-cursor-end ABC))))

  (test-with-random-fixlen-string s 32
    (test s
          (string-append (substring/cursors s 0 8)
                         (substring/cursors s 8 20)
                         (substring/cursors s 20 32)))
    )
  (test #t (type-exception (substring/cursors #t 0 0)))
  (test #t (type-exception (substring/cursors "" 8.2 0)))
  (test #t (type-exception (substring/cursors "" 0 8.2)))
  )

(test-group "string-copy/cursors"
  (test "" (string-copy/cursors ""))

  (test-with-random-fixlen-string s 10
    (test #f (eqv? s (string-copy/cursors s)))
    (test s (string-copy/cursors s))
    (test "" (string-copy/cursors s (string-cursor-end s)))
    (test #f (eqv? s (string-copy/cursors s 9)))
    (test (utf8:substring s 9) (string-copy/cursors s 9))
    (test s (string-copy/cursors s (string-cursor-start s)))
    (test s (string-copy/cursors s
                                 (string-cursor-start s)
                                 (string-cursor-end s)))
    (test (utf8:substring s 1 2) (string-copy/cursors s 1 2))
    (test "" (string-copy/cursors s 0 0))
    (test "" (string-copy/cursors s 10 10)))
  (test #t (type-exception (string-copy/cursors #t)))
  (test #t (type-exception (string-copy/cursors "" 0.2)))
  (test #t (type-exception (string-copy/cursors "" 0 2.2)))
  )

(test-group "string-take"
  (test "" (string-take "" 0))

  (test-with-random-string s
    (let* ((len (utf8:string-length s))
           (n (quotient len 2)))
      (test "" (string-take s 0))
      (test s (string-take s len))
      (test (utf8:substring s 0 n) (string-take s n))
    ))
  (test #t (type-exception (string-take 0 0)))
  (test #t (type-exception (string-take "" 0.2)))
  )

(test-group "string-drop"
  (test "" (string-drop "" 0))

  (test-with-random-string s
    (let* ((len (utf8:string-length s))
           (n (quotient len 2)))
      (test s (string-drop s 0))
      (test "" (string-drop s len))
      (test (utf8:substring s n) (string-drop s n))
    ))
  (test #t (type-exception (string-drop 0 0)))
  (test #t (type-exception (string-drop "" 0.2)))
  )

(test-group "string-take-right"
  (test "" (string-take-right "" 0))

  (test-with-random-string s
    (let* ((len (utf8:string-length s))
           (n (quotient len 2)))
      (test "" (string-take-right s 0))
      (test s (string-take-right s len))
      (test (utf8:substring s (- len n)) (string-take-right s n))
    ))
  (test #t (type-exception (string-take-right 0 0)))
  (test #t (type-exception (string-take-right "" 0.2)))
  )

(test-group "string-drop-right"
  (test "" (string-drop-right "" 0))

  (test-with-random-string s
    (let* ((len (utf8:string-length s))
           (n (quotient len 2)))
      (test s (string-drop-right s 0))
      (test "" (string-drop-right s len))
      (test (utf8:substring s 0 (- len n)) (string-drop-right s n))
    ))
  (test #t (type-exception (string-drop-right 0 0)))
  (test #t (type-exception (string-drop-right "" 0.2)))
  )

(test-group "string-pad"
  (test "" (string-pad "" 0))
  (test "*****" (string-pad "" 5 #\*))

  (test-with-random-fixlen-string s 10
    (test 8 (utf8:string-length (string-pad s 8)))
    (test 14 (utf8:string-length (string-pad s 14)))
    (test (utf8:substring s 6) (string-pad s 4))
    (test (string-append "  " s) (string-pad s 12))
    (test (string-append "aa" s) (string-pad s 12 #\a))
    (test (string-append "αα" s) (string-pad s 12 (integer->char #x3b1)))
    (test (string-append "aa" (utf8:substring s 6))
          (string-pad s 6 #\a 6))
    (test (string-append "aa" (utf8:substring s 4 7))
          (string-pad s 5 #\a 4 7))
    )
  (test #t (type-exception (string-pad 0 10 #\*)))
  (test #t (type-exception (string-pad "" 8.2 #\*)))
  (test #t (type-exception (string-pad "" 8 "hello")))
  (test #t (type-exception (string-pad "" 8 #\* #t)))
  (test #t (type-exception (string-pad "" 8 #\* 0 #t)))
  )

(test-group "string-pad-right"
  (test "" (string-pad-right "" 0))
  (test "*****" (string-pad-right "" 5 #\*))

  (test-with-random-fixlen-string s 10
    (test 8 (utf8:string-length (string-pad-right s 8)))
    (test 14 (utf8:string-length (string-pad-right s 14)))
    (test (utf8:substring s 0 4) (string-pad-right s 4))
    (test (string-append s "  ") (string-pad-right s 12))
    (test (string-append s "aa") (string-pad-right s 12 #\a))
    (test (string-append s "αα")
          (string-pad-right s 12 (integer->char #x3b1)))
    (test (string-append (utf8:substring s 6) "aa")
          (string-pad-right s 6 #\a 6))
    (test (string-append (utf8:substring s 4 7) "aa")
          (string-pad-right s 5 #\a 4 7))
    )
  (test #t (type-exception (string-pad-right 0 10 #\*)))
  (test #t (type-exception (string-pad-right "" 8.2 #\*)))
  (test #t (type-exception (string-pad-right "" 8 "hello")))
  (test #t (type-exception (string-pad-right "" 8 #\* #t)))
  (test #t (type-exception (string-pad-right "" 8 #\* 0 #t)))
  )

(test-group "string-trim"
  (test "" (string-trim "" (constantly #t)))

  ;; It's a little trickier to build randomized tests for these than
  ;; it is for the pad forms.  I have to ensure that the test string
  ;; doesn't contain characters I want to trim.
  (test-with-random-fixlen-string s-base 10
    (let ((s (string-translate s-base #\space)))
        (test s (string-trim s (constantly #f)))
        (test s (string-trim (string-append "   " s) is-space?))
        (test (string-append s "   ")
              (string-trim (string-append "   " s "   ") is-space?))
        (test "" (string-trim s (constantly #t)))
        (test (string-append "a  " s)
              (string-trim (string-append "    a  " s) is-space?))
        (test (utf8:substring s 4) (string-trim s (constantly #f) 4))
        (test (utf8:substring s 4 8)
              (string-trim s (constantly #f) 4 8))
        (test ""
              (string-trim (string-append "ZZ" s "ZZ")
                           (constantly #t)
                           4
                           8))
      ))
  (test #t (type-exception (string-trim 0 char?)))
  (test #t (type-exception (string-trim "" 0)))
  (test #t (type-exception (string-trim "" char? #t)))
  (test #t (type-exception (string-trim "" char? 0 #t)))
  )

(test-group "string-trim-right"
  (test "" (string-trim-right "" (constantly #t)))

  (test-with-random-fixlen-string s-base 10
    (let ((s (string-translate s-base #\space)))
        (test s (string-trim-right s (constantly #f)))
        (test s (string-trim-right (string-append s "   ") is-space?))
        (test "" (string-trim-right s (constantly #t)))
        (test (string-append "   " s)
              (string-trim-right (string-append "   " s "   ") is-space?))
        (test (string-append s "  a")
              (string-trim-right (string-append s "  a    ") is-space?))
        (test (utf8:substring s 4) (string-trim-right s (constantly #f) 4))
        (test (utf8:substring s 4 8)
              (string-trim-right s (constantly #f) 4 8))
        (test ""
              (string-trim-right (string-append "ZZ" s "ZZ")
                           (constantly #t)
                           4
                           8))
      ))
  (test #t (type-exception (string-trim-right 0 char?)))
  (test #t (type-exception (string-trim-right "" 0)))
  (test #t (type-exception (string-trim-right "" char? #t)))
  (test #t (type-exception (string-trim-right "" char? 0 #t)))
  )

(test-group "string-trim-both"
  (test-with-random-fixlen-string s-base 10
    (let ((s (string-translate s-base #\space)))
        (test s (string-trim-both s (constantly #f)))
        (test s (string-trim-both (string-append s "   ") is-space?))
        (test "" (string-trim-both s (constantly #t)))
        (test s (string-trim-both (string-append "   " s "   ") is-space?))
        (test (string-append "a  " s)
              (string-trim-both (string-append "    a  " s) is-space?))
        (test (string-append s "  a")
              (string-trim-both (string-append s "  a    ") is-space?))
        (test (utf8:substring s 4) (string-trim-both s (constantly #f) 4))
        (test (utf8:substring s 4 8)
              (string-trim-both s (constantly #f) 4 8))
        (test ""
              (string-trim-both (string-append "ZZ" s "ZZ")
                                (constantly #t)
                                4
                                8))
      ))
  (test #t (type-exception (string-trim-both 0 char?)))
  (test #t (type-exception (string-trim-both "" 0)))
  (test #t (type-exception (string-trim-both "" char? #t)))
  (test #t (type-exception (string-trim-both "" char? 0 #t)))
  )

(test-group "string-prefix-length"
  (test 0 (string-prefix-length "" ""))

  (test-with-random-string s
    (let ((s-len (utf8:string-length s)))
      (test 0 (string-prefix-length "" s))
      (test s-len (string-prefix-length s s))
      (let* ((n (quotient s-len 2))
             (s* (utf8:substring s 0 n)))
        (test n (string-prefix-length s* s))
        (test n (string-prefix-length s s*)))
      (test 0 (string-prefix-length
               (string-append (utf8:string (integer->char #x1d400)) s)
               s))
      ))
  ;; With start/end indices.
  (test-with-random-fixlen-string s 32
    (test 0 (string-prefix-length "" s 0))
    (test 0 (string-prefix-length s "" 0))
    (test 0 (string-prefix-length "" s 0 0))
    (test 0 (string-prefix-length s "" 0 20))
    (test 8 (string-prefix-length s s 0 8))
    (test 0 (string-prefix-length s "" 1 8))
    (test 0 (string-prefix-length "" s 0 0 0))
    (test 0 (string-prefix-length s "" 0 8 0))
    (test 8 (string-prefix-length s s 0 8 0))
    (test 0 (string-prefix-length "" s 0 0 0 0))
    (test 0 (string-prefix-length s "" 0 4 0 0))
    (test 8 (string-prefix-length s s 1 9 1 9))
    )
  (test #t (type-exception (string-prefix-length #t "")))
  (test #t (type-exception (string-prefix-length "" #t)))
  (test #t (type-exception (string-prefix-length "" "" 0.2)))
  (test #t (type-exception (string-prefix-length "" "" 0 #t)))
  (test #t (type-exception (string-prefix-length "" "" 0 0 'a)))
  (test #t (type-exception (string-prefix-length "" "" 0 0 0 5/3)))
  )

(test-group "string-prefix?"
  (test #t (string-prefix? "" ""))

  (test-with-random-string s
    (let ((s-len (utf8:string-length s)))
      (test #t (string-prefix? "" s))
      (test #t (string-prefix? s s))
      (let* ((n (quotient s-len 2))
             (s* (utf8:substring s 0 n)))
        (test #t (string-prefix? s* s))
        (test #t (or (equal? s s*) (not (string-prefix? s s*)))))
      (test #f (string-prefix?
                (string-append (utf8:string (integer->char #x1d400)) s)
                s))
      ))
  ;; With start/end indices.
  (test-with-random-fixlen-string s 32
    (test #t (string-prefix? "" s 0))
    (test #f (string-prefix? s "" 0))
    (test #t (string-prefix? "" s 0))
    (test #f (string-prefix? s "" 0 20))
    (test #t (string-prefix? s s 0 8))
    (test #f (string-prefix? s "" 1 8))
    (test #t (string-prefix? "" s 0 0 0))
    (test #f (string-prefix? s "" 0 8 0))
    (test #t (string-prefix? s s 0 8 0))
    (test #t (string-prefix? "" s 0 0 0 0))
    (test #f (string-prefix? s "" 0 4 0 0))
    (test #t (string-prefix? s s 1 9 1 9))
    )
  (test #t (type-exception (string-prefix? #t "")))
  (test #t (type-exception (string-prefix? "" #t)))
  (test #t (type-exception (string-prefix? "" "" 0.2)))
  (test #t (type-exception (string-prefix? "" "" 0 #t)))
  (test #t (type-exception (string-prefix? "" "" 0 0 'a)))
  (test #t (type-exception (string-prefix? "" "" 0 0 0 5/3)))
  )

(test-group "string-suffix-length"
  (test 0 (string-suffix-length "" ""))

  (test-with-random-string s
    (let ((s-len (utf8:string-length s)))
      (test 0 (string-suffix-length "" s))
      (test s-len (string-suffix-length s s))
      (let* ((n (quotient s-len 2))
             (m (- s-len n))
             (s* (utf8:substring s n)))
        (test m (string-suffix-length s* s))
        (test m (string-suffix-length s s*)))
      (test 0 (string-suffix-length
               (string-append s (utf8:string (integer->char #x1d400)))
               s))
      ))

  ;; With start/end indices.
  (test-with-random-fixlen-string s 32
    (test 0 (string-suffix-length "" s 0))
    (test 0 (string-suffix-length s "" 0))
    (test 0 (string-suffix-length "" s 0 0))
    (test 0 (string-suffix-length s "" 0 20))
    (test 8 (string-suffix-length s s 24 32))
    (test 0 (string-suffix-length s "" 25 32))
    (test 0 (string-suffix-length "" s 0 0 0))
    (test 0 (string-suffix-length s "" 24 32 0))
    (test 8 (string-suffix-length s s 24 32 0))
    (test 0 (string-suffix-length "" s 0 0 0 0))
    (test 0 (string-suffix-length s "" 28 32 0 0))
    (test 8 (string-suffix-length s s 23 31 23 31))
    )
  (test #t (type-exception (string-suffix-length #t "")))
  (test #t (type-exception (string-suffix-length "" #t)))
  (test #t (type-exception (string-suffix-length "" "" 0.2)))
  (test #t (type-exception (string-suffix-length "" "" 0 #t)))
  (test #t (type-exception (string-suffix-length "" "" 0 0 'a)))
  (test #t (type-exception (string-suffix-length "" "" 0 0 0 5/3)))
  )

(test-group "string-suffix?"
  (test #t (string-suffix? "" ""))

  (test-with-random-string s
    (let ((s-len (utf8:string-length s)))
      (test #t (string-suffix? "" s))
      (test #t (string-suffix? s s))
      (let* ((n (quotient s-len 2))
             (m (- s-len n))
             (s* (utf8:substring s n)))
        (test #t (string-suffix? s* s))
        (test #t (or (equal? s s*) (not (string-suffix? s s*)))))
      (test #f (string-suffix?
                (string-append s (utf8:string (integer->char #x1d400)))
                s))
      ))

  ;; With start/end indices.
  (test-with-random-fixlen-string s 32
    (test #t (string-suffix? "" s 0))
    (test #f (string-suffix? s "" 0))
    (test #t (string-suffix? "" s 0))
    (test #f (string-suffix? s "" 0 20))
    (test #t (string-suffix? s s 24 32))
    (test #f (string-suffix? s "" 25 32))
    (test #t (string-suffix? "" s 0 0 0))
    (test #f (string-suffix? s "" 24 32 0))
    (test #t (string-suffix? s s 24 32 0))
    (test #t (string-suffix? "" s 0 0 0 0))
    (test #f (string-suffix? s "" 28 32 0 0))
    (test #t (string-suffix? s s 23 31 23 31))
    )
  (test #t (type-exception (string-suffix? #t "")))
  (test #t (type-exception (string-suffix? "" #t)))
  (test #t (type-exception (string-suffix? "" "" 0.2)))
  (test #t (type-exception (string-suffix? "" "" 0 #t)))
  (test #t (type-exception (string-suffix? "" "" 0 0 'a)))
  (test #t (type-exception (string-suffix? "" "" 0 0 0 5/3)))
  )

;;;; Searching

(test-group "string-index"
  (test #t (string-cursor=? (string-cursor-start "")
                            (string-index "" char?)))
  (test #t (string-cursor=?
            (string-index->cursor ABC 1)
            (string-index ABC (cut char=? <> (integer->char #x3b2)))))
  (test #t (string-cursor=?
            (string-cursor-start ascii)
            (string-index ascii char-upper-case?)))
  (test #t (string-cursor=?
            (string-index->cursor ABCDEF 1)
            (string-index ABCDEF char-lower-case?)))

  (test-with-random-string s-base
    (let ((s (string-translate s-base #\space)))
      (test #t (string-cursor=? (string-cursor-start s)
                                (string-index s (constantly #t))))
      (test #t (string-cursor=? (string-cursor-end s)
                                (string-index s (constantly #f))))
      (let ((s* (string-append " " s " ")))
        (test #t
              (string-cursor=? (string-cursor-start s*)
                               (string-index s* is-space?))))
      ))
  (test #t (type-exception (string-index 0 char?)))
  (test #t (type-exception (string-index "" 0)))
  (test #t (type-exception (string-index "" char? 0.2)))
  (test #t (type-exception (string-index "" char? 0 0.2)))
  )
 
(test-group "string-index-right"
  (test #t (string-cursor=? (string-cursor-start "")
                            (string-index-right "" char?)))
  (test #t (string-cursor=?
            (string-index->cursor ABC 1)
            (string-index-right ABC (cut char=? <> (integer->char #x3b2)))))
  (test #t (string-cursor=?
            (string-cursor-prev ascii (string-cursor-end ascii))
            (string-index-right ascii char-upper-case?)))
  (test #t (string-cursor=?
            (string-cursor-prev ABCDEF (string-cursor-end ABCDEF))
            (string-index-right ABCDEF char-lower-case?)))

  (test-with-random-string s-base
    (let ((s (string-translate s-base #\space)))
      (test #t (string-cursor=?
                (if (equal? s "")
                    (string-cursor-end s)
                    (string-cursor-prev s (string-cursor-end s)))
                (string-index-right s (constantly #t))))
      (test #t (string-cursor=? (string-cursor-start s)
                                (string-index-right s (constantly #f))))
      (let ((s* (string-append " " s " ")))
        (test #t
              (string-cursor=? (string-cursor-prev s* (string-cursor-end s*))
                               (string-index-right s* is-space?))))
      ))
  (test #t (type-exception (string-index-right 0 char?)))
  (test #t (type-exception (string-index-right "" 0)))
  (test #t (type-exception (string-index-right "" char? 0.2)))
  (test #t (type-exception (string-index-right "" char? 0 0.2)))
  )

(test-group "string-skip"
  (test #t (string-cursor=? (string-cursor-start "")
                            (string-skip "" char?)))
  (test #t (string-cursor=?
            (string-index->cursor ABC 1)
            (string-skip ABC (cut char=? <> (integer->char #x3b1)))))
  (test #t (string-cursor=?
            (string-cursor-start ascii)
            (string-skip ascii char-lower-case?)))
  (test #t (string-cursor=?
            (string-cursor-start ABCDEF)
            (string-skip ABCDEF char-lower-case?)))

  (test-with-random-string s-base
    (let ((s (string-translate s-base #\space)))
      (test #t (string-cursor=? (string-cursor-start s)
                                (string-skip s (constantly #f))))
      (test #t (string-cursor=? (string-cursor-end s)
                                (string-skip s (constantly #t))))
      (let ((s* (string-append " " s " ")))
        (test #t
              (string-cursor=? (if (equal? s "")
                                   (string-cursor-end s*)
                                   (string-index->cursor s* 1))
                               (string-skip s* is-space?))))
      ))
  (test #t (type-exception (string-skip 0 char?)))
  (test #t (type-exception (string-skip "" 0)))
  (test #t (type-exception (string-skip "" char? 0.2)))
  (test #t (type-exception (string-skip "" char? 0 0.2)))
  )

(test-group "string-skip-right"
  (test #t (string-cursor=? (string-cursor-start "")
                            (string-skip-right "" char?)))
  (test #t (string-cursor=?
            (string-index->cursor ABC 1)
            (string-skip-right ABC (cut char=? <> (integer->char #x3b3)))))
  (test #t (string-cursor=?
            (string-cursor-prev ascii (string-cursor-end ascii))
            (string-skip-right ascii char-lower-case?)))
  (test #t (string-cursor=?
            (string-index->cursor ABCDEF 4)
            (string-skip-right ABCDEF char-lower-case?)))

  (test-with-random-string s-base
    (let ((s (string-translate s-base #\space)))
      (test #t (string-cursor=?
                (if (equal? s "")
                    (string-cursor-end s)
                    (string-cursor-prev s (string-cursor-end s)))
                (string-skip-right s (constantly #f))))
      (test #t (string-cursor=? (string-cursor-start s)
                                (string-skip-right s (constantly #t))))
      (let ((s* (string-append " " s " ")))
        (test #t
              (string-cursor=? (if (equal? s "")
                                   (string-cursor-start s*)
                                   (string-cursor-back s*
                                                       (string-cursor-end s*)
                                                       2))
                               (string-skip-right s* is-space?))))
      ))
  (test #t (type-exception (string-skip-right 0 char?)))
  (test #t (type-exception (string-skip-right "" 0)))
  (test #t (type-exception (string-skip-right "" char? 0.2)))
  (test #t (type-exception (string-skip-right "" char? 0 0.2)))
  )

(test-group "string-contains"
  (test 0 (string-cursor->index "" (string-contains "" "")))
  (test 0 (string-cursor->index "" (string-contains "" "" 0)))
  (test 0 (string-cursor->index "" (string-contains "" "" 0 0)))
  (test 0 (string-cursor->index "" (string-contains "" "" 0 0 0)))
  (test 0 (string-cursor->index "" (string-contains "" "" 0 0 0 0)))

  (let ((s "áƀçďěƒƒƒƒőô"))
    (test 0 (string-cursor->index s (string-contains s "á")))
    (test 5 (string-cursor->index s (string-contains s "ƒƒ")))
    (test 4 (string-cursor->index s (string-contains s "ěƒƒ")))
    (test 8 (string-cursor->index s (string-contains s "ƒőô")))
    (test #f (string-contains s "ěƒƒƒőô"))
    (test 0 (string-cursor->index s (string-contains s "á" 0)))
    (test #f (string-contains s "á" 1))
    (test 5 (string-cursor->index s (string-contains s "ƒƒ" 1)))
    (test 4 (string-cursor->index s (string-contains s "ěƒƒ" 3 10)))
    (test #f (string-contains s "ƒőô" 3 10))
    (test 3 (string-cursor->index s (string-contains s "ďěƒ" 3 10 0)))
    (test 3 (string-cursor->index s (string-contains s "soďěƒ" 3 10 2)))
    (test #f (string-contains s "soďěƒ" 3 10 1))
    (test 5 (string-cursor->index s
                                  (string-contains s "boƒƒer" 1 10 2 4)))
    (test #f (string-contains s "boƒƒer" 1 10 2 5))
    )

  (test-with-random-string s-base
    (let ((s (string-translate s-base #\a)))
      (test 0 (string-cursor->index s (string-contains s "")))
      (test 0 (string-cursor->index s (string-contains s s)))
      (test #f (string-contains s "a"))

      (let ((s* (string-append s (make-string 8 #\a))))
        (test 0 (string-cursor->index s* (string-contains s* s)))
        (test (utf8:string-length s)
              (string-cursor->index s* (string-contains s* "a"))))

      (let ((s* (string-append (make-string 8 #\a) s)))
        (test #t (or (equal? s "")
                     (= 8
                        (string-cursor->index s*
                                              (string-contains s* s)))))
        (test 0 (string-cursor->index s* (string-contains s* "a"))))
    ))
  (test #t (type-exception (string-contains #t "")))
  (test #t (type-exception (string-contains "" #t)))
  (test #t (type-exception (string-contains "" "" 0.2)))
  (test #t (type-exception (string-contains "" "" 0 #t)))
  (test #t (type-exception (string-contains "" "" 0 0 'a)))
  (test #t (type-exception (string-contains "" "" 0 0 0 5/3)))
  )

(test-group "string-contains-right"
  (test 0 (string-cursor->index "" (string-contains-right "" "")))
  (test 0 (string-cursor->index "" (string-contains-right "" "" 0)))
  (test 0 (string-cursor->index "" (string-contains-right "" "" 0 0)))
  (test 0 (string-cursor->index "" (string-contains-right "" "" 0 0 0)))
  (test 0 (string-cursor->index "" (string-contains-right "" "" 0 0 0 0)))

  (let ((s "áƀçďěƒƒƒƒőô"))
    (test 0 (string-cursor->index s (string-contains-right s "á")))
    (test 7 (string-cursor->index s (string-contains-right s "ƒƒ")))
    (test 4 (string-cursor->index s (string-contains-right s "ěƒƒ")))
    (test 8 (string-cursor->index s (string-contains-right s "ƒőô")))
    (test #f (string-contains-right s "ěƒƒƒőô"))
    (test 0 (string-cursor->index s (string-contains-right s "á" 0)))
    (test #f (string-contains-right s "á" 1))
    (test 7 (string-cursor->index s (string-contains-right s "ƒƒ" 1)))
    (test 4 (string-cursor->index s (string-contains-right s "ěƒƒ" 3 10)))
    (test #f (string-contains-right s "ƒőô" 3 10))
    (test 3 (string-cursor->index s (string-contains-right s "ďěƒ" 3 10 0)))
    (test 3 (string-cursor->index s (string-contains-right s "soďěƒ" 3 10 2)))
    (test #f (string-contains-right s "soďěƒ" 3 10 1))
    (test 7 (string-cursor->index s
                                  (string-contains-right s "boƒƒer" 1 10 2 4)))
    (test #f (string-contains-right s "boƒƒer" 1 10 2 5))
    )

  (test-with-random-string s-base
    (let ((s (string-translate s-base #\a)))
      (test (utf8:string-length s)
            (string-cursor->index s (string-contains-right s "")))
      (test 0 (string-cursor->index s (string-contains-right s s)))
      (test #f (string-contains-right s "a"))

      (let ((s* (string-append s (make-string 8 #\a))))
        (test #t
              (if (equal? s "")
                  (= 8 (string-cursor->index
                        s*
                        (string-contains-right s* s)))
                  (= 0 (string-cursor->index
                        s*
                        (string-contains-right s* s)))))
        (test (- (utf8:string-length s*) 1)
              (string-cursor->index s* (string-contains-right s* "a")))
        (test (utf8:string-length s)
              (string-cursor->index
               s*
               (string-contains-right s* (make-string 8 #\a)))))

      (let ((s* (string-append (make-string 8 #\a) s)))
        (test #t (or (equal? s "")
                     (= 8
                        (string-cursor->index s*
                                              (string-contains-right s* s)))))
        (test 7 (string-cursor->index s* (string-contains-right s* "a"))))
    ))
  (test #t (type-exception (string-contains-right #t "")))
  (test #t (type-exception (string-contains-right "" #t)))
  (test #t (type-exception (string-contains-right "" "" 0.2)))
  (test #t (type-exception (string-contains-right "" "" 0 #t)))
  (test #t (type-exception (string-contains-right "" "" 0 0 'a)))
  (test #t (type-exception (string-contains-right "" "" 0 0 0 5/3)))
  )

;;;; The whole string

(test-group "string-reverse"
  (test "" (string-reverse ""))
  (test "ƒěďçƀá" (string-reverse "áƀçďěƒ"))
  (test "" (string-reverse "" (string-cursor-start "")))
  (test "ƒěďçƀá" (string-reverse "áƀçďěƒ" 0))
  (test "ƒěďç" (string-reverse "áƀçďěƒ" 2))
  (test "" (string-reverse "" 0 (string-cursor-end "")))
  (test "ƒěďçƀá" (string-reverse "áƀçďěƒ" 0 6))
  (test "ěďç" (string-reverse "áƀçďěƒ" 2 5))
  (test #t (type-exception (string-reverse 0)))
  (test #t (type-exception (string-reverse "" 0.2)))
  (test #t (type-exception (string-reverse "" 0 0.2)))
  )

(test-group "string-concatenate"
  (test "" (string-concatenate '()))

  (test-generative ((s1 (lambda ()
                          (random-string test-string-max-length)))
                    (s2 (lambda ()
                          (random-string test-string-max-length))))
    (test (string-append s1 s2)
          (string-concatenate (list "" s1 "" s2 "" "")))
    (test (string-append s2 s1)
          (string-concatenate (list "" "" s2 "" s1 "")))
    )
  (test #t (type-exception (string-concatenate #t)))
  )

(test-group "string-concatenate-reverse"
  (test "" (string-concatenate-reverse '()))

  (test-generative ((s1 (lambda ()
                          (random-string test-string-max-length)))
                    (s2 (lambda ()
                          (random-string test-string-max-length))))
    (test s1 (string-concatenate-reverse '() s1))
    (test (string-append s2 s1)
          (string-concatenate-reverse (list "" s1 "" s2 "" "")))
    (test (string-append s1 s2)
          (string-concatenate-reverse (list "" "" s2 "" s1 "")))
    (test (string-append s1 s2 "xy")
          (string-concatenate-reverse (list "" "" s2 "" s1 "") "xy"))
    (test (string-append s2 s1 "huh?")
          (string-concatenate-reverse (list "" s1 "" s2 "" "")
                                      "huh?"
                                      (string-cursor-end "huh?")))
    (test (string-append s2 s1 "huh")
          (string-concatenate-reverse (list "" s1 "" s2 "" "")
                                      "huh?"
                                      3))
    )
  (test #t (type-exception (string-concatenate-reverse #t)))
  (test #t (type-exception (string-concatenate-reverse '() #t)))
  (test #t (type-exception (string-concatenate-reverse '() "" 0.2)))
  )

(test-group "string-fold"
  (test 'x (string-fold cons 'x ""))
  (test-with-random-string s
    (let ((lis (reverse (utf8:string->list s))))
      (test lis (string-fold cons '() s))
      (test (fold + 0 (map char->integer lis))
            (string-fold (lambda (c sum) (+ (char->integer c) sum))
                         0
                         s))
      (test (count unicode-char-printing? lis)
            (string-fold (lambda (c count)
                           (if (unicode-char-printing? c)
                               (+ count 1)
                               count))
                         0
                         s))))
  (test #t (type-exception (string-fold #t 0 "")))
  (test #t (type-exception (string-fold cons '() 0)))
  )

(test-group "string-fold-right"
  (test 'x (string-fold-right cons 'x ""))
  (test-with-random-string s
    (let ((lis (utf8:string->list s)))
      (test lis (string-fold-right cons '() s))
      (test (fold + 0 (map char->integer lis))
            (string-fold (lambda (c sum) (+ (char->integer c) sum))
                         0
                         s))
      (test (count unicode-char-printing? lis)
            (string-fold-right (lambda (c count)
                           (if (unicode-char-printing? c)
                               (+ count 1)
                               count))
                         0
                         s))))
  (test #t (type-exception (string-fold-right #t 0 "")))
  (test #t (type-exception (string-fold-right cons '() 0)))
  )

(test-group "string-for-each-cursor"
  (test-with-random-string s
    (test (reverse (utf8:string->list s))
          (let ((res '()))
            (string-for-each-cursor
             (lambda (cur)
               (set! res (cons (string-ref/cursor s cur) res)))
             s)
           res)))
  (test #t (type-exception (string-for-each-cursor #t "")))
  (test #t (type-exception (string-for-each-cursor values #t)))
  )

(test-group "string-replicate"
  (test "" (string-replicate "" 0 0))
  (test "ƀçďěƒá" (string-replicate "áƀçďěƒ" 1 7))
  (test "ƒáƀçďě" (string-replicate "áƀçďěƒ" -1 5))
  (test "çďěƒáƀçďěƒáƀçď" (string-replicate "áƀçďěƒ" -4 10))
  (test "ƀçďěƒƀçďěƒƀçď" (string-replicate "áƀçďěƒ" 90 103 1))
  (test "ěçďěçďěçďě" (string-replicate "áƀçďěƒ" -13 -3 2 5))

  (test-with-random-string s
    (let* ((s-len (utf8:string-length s))
           (k (quotient s-len 2))
           (k-sub (utf8:substring s 0 k)))
      (test "" (string-replicate s 0 0))
      (test "" (string-replicate s 0 0 0))
      (test "" (string-replicate s (* s-len 1000) (* s-len 1000)))
      (test s (string-replicate s 0 s-len))
      (test s (string-replicate s 0 s-len 0 s-len))
      (test s (string-replicate s (- s-len) 0))
      (test s (string-replicate s (* s-len 1000) (* s-len 1001)))
      (test (string-append s s) (string-replicate s (- s-len) s-len))
      (test k-sub (string-replicate s 0 k))
      (test (string-append k-sub k-sub)
            (string-replicate s 0 (* k 2) 0 k))
      (test (utf8:substring s (- s-len k))
            (string-replicate s (- k) 0))
      ))
  (test #t (type-exception (string-replicate "" 0 5/3)))
  (test #t (type-exception (string-replicate "" 0 -3 0.2)))
  (test #t (type-exception (string-replicate "" 0 -3 0 0.2)))
  )

(test-group "string-count"
  (test 0 (string-count "" char?))

  (test-with-random-string s-base
    (let* ((s (string-translate s-base #\space))
           (k (quotient (utf8:string-length s) 2)))
      (test (utf8:string-length s)
            (string-count s (complement is-space?)))
      (test 0 (string-count s (constantly #f)))
      (test 2 (string-count (string-append "  " s) is-space?))
      (test 2 (string-count (string-append s "  ") is-space?))
      (test 4 (string-count (string-append "  " s "  ") is-space?))
      (test k (string-count s char? 0 k))
      (test (- (utf8:string-length s) k) (string-count s char? k))
      (test 0 (string-count s (constantly #f) 0 k))
      ))
  (test #t (type-exception (string-count #t char?)))
  (test #t (type-exception (string-count "" #t)))
  (test #t (type-exception (string-count "" char? 0.2)))
  (test #t (type-exception (string-count "" 0 5/3)))
  )

(test-group "string-replace"
  (test "" (string-replace "" "" 0 0))

  (test-with-random-string s
    (let* ((s-len (utf8:string-length s))
           (abcdef-len (utf8:string-length ABCDEF))
           (k (pseudo-random-integer s-len))
           (l (pseudo-random-integer abcdef-len)))
      (test "" (string-replace s "" (string-cursor-start s) s-len))
      (test s (string-replace s "" 0 (string-cursor-start s)))
      (test (utf8:substring s 0 k)
            (string-replace s "" k (string-cursor-end s)))
      (test (string-append ABCDEF s) (string-replace s ABCDEF 0 0))
      (test (string-append ABCDEF s)
            (string-replace s ABCDEF 0 0 (string-cursor-start ABCDEF)))
      (test (string-append s ABCDEF)
            (string-replace s ABCDEF s-len s-len))
      (test (string-append s ABCDEF)
            (string-replace s ABCDEF s-len s-len 0 abcdef-len))
      (test (string-append (utf8:substring s 0 k)
                           ABCDEF
                           (utf8:substring s k))
            (string-replace s ABCDEF k k))
      (test (string-append (utf8:substring s 0 k)
                           (utf8:substring ABCDEF l)
                           (utf8:substring s k))
            (string-replace s ABCDEF k k l))
      (test (string-append (utf8:substring s 0 k)
                           (utf8:substring ABCDEF 0 l)
                           (utf8:substring s k))
            (string-replace s ABCDEF k k 0 l))

      (let ((m (min (+ k 1) s-len)))
        (test (string-append (utf8:substring s 0 k) (utf8:substring s m))
              (string-replace s "" k m))
        (test (string-append (utf8:substring s 0 k)
                             ABCDEF
                             (utf8:substring s m s-len))
              (string-replace s ABCDEF k m))
        (test (string-append (utf8:substring s 0 k)
                             (utf8:substring ABCDEF l)
                             (utf8:substring s m s-len))
              (string-replace s ABCDEF k m l))
        (test (string-append (utf8:substring s 0 k)
                             (utf8:substring ABCDEF 0 l)
                             (utf8:substring s m s-len))
              (string-replace s ABCDEF k m 0 l)))
      ))
  (test #t (type-exception (string-replace #t "" 0 0)))
  (test #t (type-exception (string-replace "" #t 0 0)))
  (test #t (type-exception (string-replace "" "" 0.2 0)))
  (test #t (type-exception (string-replace "" "" 0 0.2)))
  (test #t (type-exception (string-replace "" "" 0 0 0.2)))
  (test #t (type-exception (string-replace "" "" 0 0 0 0.2)))
  )

(test-group "string-split"
  (test '() (string-split "" ""))

  (test-with-random-string s-base
    (let* ((s (string-translate s-base #\space))
           (s-len (utf8:string-length s))
           (k (pseudo-random-integer s-len)))
      (test (map utf8:string (utf8:string->list s))
            (string-split s ""))
      (test (list s) (string-split s " " 'infix))
      (test #t (raises-exception (string-split s " " 'uffix)))
      (test #t (raises-exception (string-split s "" "" 'strict-infix)))
      (test (list "" s) (string-split (string-append " " s) " " 'infix))
      (test (list s) (string-split (string-append " " s) " " 'prefix))
      (test (list s) (string-split (string-append s " ") " " 'suffix))
      (test (list s "" s)
            (string-split (string-append s "  " s) " " 'infix #f))
      (test (list s "" s "")
            (string-split (string-append " " s "  " s " ") " " 'prefix #f))
      (test (list s s s "")
            (string-split (string-append "   " s "   " s "   " s "   ")
                          "   "
                          'prefix))
      (test (list "" s s s)
            (string-split (string-append "   " s "   " s "   " s "   ")
                          "   "
                          'suffix))
      (test (list s "" (string-append " " s))
            (string-split (string-append s "   " s) " " 'infix 2))
      (test (list "" s s)
            (string-split (string-append " " s " " s) " " 'strict-infix 3))
      (test (list "" s s " a")
            (string-split (string-append " " s " " s "  a")
                          " "
                          'strict-infix
                          3))
      (test (list s s)
            (string-split (string-append " " s " " s) " " 'prefix 3))
      (test (list s s " a")
            (string-split (string-append " " s " " s "  a") " " 'prefix 3))
      (test (list (utf8:substring s k)) (string-split s " " 'infix #f k))
      (test (list (utf8:substring s 0 k))
            (string-split s " " 'infix #f 0 k))
      (test (take (map utf8:string (utf8:string->list s)) k)
            (string-split s "" 'infix #f 0 k))
      (test (list s)
            (string-split (string-append " " s " ")
                          " "
                          'infix
                          #f
                          1
                          (+ s-len 1)))
    ))
  (test #t (type-exception (string-split #t "")))
  (test #t (type-exception (string-split "" #t)))
  (test #t (type-exception (string-split "" "" #t)))
  (test #t (type-exception (string-split "" "" 'infix 3.1)))
  (test #t (type-exception (string-split "" "" 'infix 3 'b)))
  )

(test-group "string-filter"
  (test "ABC" (string-filter (lambda (c)
                               (memv c (utf8:string->list "CAB")))
                             ascii))
  (test "αβ" (string-filter (lambda (c)
                              (memv c (utf8:string->list "βα")))
                            ABC))
  (test "ÀÇÉ" (string-filter (lambda (c)
                               (memv c (utf8:string->list "ÇÉÀ")))
                             ABCDEF))
  (test "BC" (string-filter (lambda (c)
                              (memv c (utf8:string->list "CAB")))
                            ascii
                            1
                            5))
  (test "β" (string-filter (lambda (c)
                             (memv c (utf8:string->list "βα")))
                           ABC
                           1))
  (test "ÇÉ" (string-filter (lambda (c)
                              (memv c (utf8:string->list "ÇÉÀ")))
                            ABCDEF
                            2
                            5))

  (test-with-random-string s-base
    (let* ((s (string-translate s-base #\space))
           (k (quotient (utf8:string-length s) 2)))
      (test s (string-filter (constantly #t) s))
      (test "" (string-filter (constantly #f) s))
      (test "    " (string-filter is-space? (string-append "  " s "  ")))
      (test s (string-filter (constantly #t) s (string-cursor-start s)))
      (test (utf8:substring s 0 k) (string-filter (constantly #t) s 0 k))
      (test (utf8:substring s k)
            (string-filter (constantly #t) s k (string-cursor-end s)))
      ))
  (test #t (type-exception (string-filter #t "")))
  (test #t (type-exception (string-filter char? #t)))
  (test #t (type-exception (string-filter char? "" 0.2)))
  (test #t (type-exception (string-filter char? "" 0 0.2)))
  )

(test-group "string-remove"
  (test "DEF" (string-remove (lambda (c)
                               (memv c (utf8:string->list "CAB")))
                             ascii))
  (test "γ" (string-remove (lambda (c)
                              (memv c (utf8:string->list "βα")))
                           ABC))
  (test "bdf" (string-remove (lambda (c)
                               (memv c (utf8:string->list "ÇÉÀ")))
                             ABCDEF))
  (test "DE" (string-remove (lambda (c)
                              (memv c (utf8:string->list "CAB")))
                            ascii
                            1
                            5))
  (test "γ" (string-remove (lambda (c)
                              (memv c (utf8:string->list "βα")))
                           ABC
                           1))
  (test "d" (string-remove (lambda (c)
                             (memv c (utf8:string->list "ÇÉÀ")))
                           ABCDEF
                           2
                           5))

  (test-with-random-string s-base
    (let* ((s (string-translate s-base #\space))
           (k (quotient (utf8:string-length s) 2)))
      (test "" (string-remove (constantly #t) s))
      (test s (string-remove (constantly #f) s))
      (test s (string-remove is-space? (string-append "  " s "  ")))
      (test "" (string-remove (constantly #t) s (string-cursor-start s)))
      (test (utf8:substring s 0 k) (string-remove (constantly #f) s 0 k))
      (test (utf8:substring s k)
            (string-remove (constantly #f) s k (string-cursor-end s)))
      ))
  (test #t (type-exception (string-remove #t "")))
  (test #t (type-exception (string-remove char? #t)))
  (test #t (type-exception (string-remove char? "" 0.2)))
  (test #t (type-exception (string-remove char? "" 0 0.2)))
  )

(test-exit)
